/* This software is contributed or developed by KYOCERA Corporation. */
/* (C) 2012 KYOCERA Corporation                                      */


#include  <linux/types.h>
#include  <linux/mtd/mtd.h>
#include  <linux/err.h>
#include  <linux/slab.h>
#include  "hs_pl_table_proc.h"
#include  "../../../arch/arm/mach-msm/smd_private.h"

#ifdef  FEATURE_KYOCERA_DEV_PLTBL

static struct mtd_info *mtd;

u_short PLTable_table_setting( void )
{
  u_short         make_ret;

  mtd = get_mtd_device(NULL, 4);
  if (IS_ERR(mtd))
  {
    printk(KERN_ERR "###PLTABLE PLTable_table_setting() 1 faild###\n");
    return FALSE;
  }
  make_ret = pltable_table_make();

  if (make_ret == FALSE)
  {
    printk(KERN_ERR "###PLTABLE PLTable_table_setting() 2 faild###\n");
  }
  return(make_ret);
}

u_long  PLTable_real_address( u_long logi_addr )
{
  u_short         physi_blkno;
  u_short         logi_blkno;
  u_long          physi_addr, offset_addr;

  logi_blkno  = logi_addr / FLASH_NAND_DEVICE_BLOCK_SIZE;
  physi_blkno = pltable_get_real(logi_blkno);

  if (physi_blkno != HS_PL_ERR_ADDRESS)
  {
    offset_addr = logi_addr % FLASH_NAND_DEVICE_BLOCK_SIZE;
    physi_addr  = (physi_blkno * FLASH_NAND_DEVICE_BLOCK_SIZE) + offset_addr;
  }
  else
  {
    printk(KERN_ERR "###PLTABLE PLTable_real_address() faild###\n");
    physi_addr = HS_PL_ERR_ADDRESS;
  }
#ifdef NO_PLTBL
  physi_addr = logi_addr;
#endif
  return(physi_addr);
}

u_long  PLTable_real_block( u_long logi_blkno )
{
  u_long          physi_blkno;
  physi_blkno = pltable_get_real(logi_blkno);
#ifdef NO_PLTBL
  physi_blkno = logi_blkno;
#endif
  return(physi_blkno);
}

u_long  PLTable_real_page( u_long logi_pno )
{
  u_long          physi_blkno;
  u_long          logi_blkno;
  u_long          physi_pno, offset_pno;
  logi_blkno  = logi_pno / FLASH_NAND_DEVICE_PAGE_COUNT;
  physi_blkno = pltable_get_real(logi_blkno);

  if (physi_blkno != HS_PL_ERR_ADDRESS)
  {
    offset_pno  = logi_pno % FLASH_NAND_DEVICE_PAGE_COUNT;
    physi_pno   = (physi_blkno * FLASH_NAND_DEVICE_PAGE_COUNT) + offset_pno;
  }
  else
  {
    printk(KERN_ERR "###PLTABLE PLTable_real_page() faild###\n");
    physi_pno = HS_PL_ERR_ADDRESS;
  }
#ifdef NO_PLTBL
  physi_pno = logi_pno;
#endif
  return(physi_pno);
}

u_long  PLTable_bad_recovery( u_long bad_addr )
{
#if 1
  u_short         chg_ret;
  u_long          chg_block;
  u_long          bad_block;
  u_long          chg_addr, offset_addr;

  if ((pltable->status1 != HS_PL_MAKE_FLAG1) ||
      (pltable->status2 != HS_PL_MAKE_FLAG2))
  {
    printk(KERN_ERR "###PLTABLE PLTable_bad_recovery() 1 faild###\n");
    chg_block = HS_PL_ERR_RECOVERY;
    return(chg_block);
  }

  bad_block = bad_addr / FLASH_NAND_DEVICE_BLOCK_SIZE;
  chg_block = pltable_search_spare(bad_block);

  if (chg_block == HS_PL_ERR_RECOVERY)
  {
    printk(KERN_ERR "###PLTABLE PLTable_bad_recovery() 2 faild###\n");
    chg_addr = HS_PL_ERR_RECOVERY;
  }
  else
  {
    offset_addr = bad_addr % FLASH_NAND_DEVICE_BLOCK_SIZE;
    chg_addr    = (chg_block * FLASH_NAND_DEVICE_BLOCK_SIZE) + offset_addr;
    chg_ret = pltable_put_real(bad_block, chg_block);
    if (chg_ret == FALSE)
    {
      pltable->status1 = (u_long)chg_ret;
      pltable->status2 = (u_long)chg_ret;
      printk(KERN_ERR "###PLTABLE PLTable_bad_recovery() 3 faild###\n");
      chg_addr = HS_PL_ERR_RECOVERY;
    }
  }

  return(chg_addr);
#else
  return(HS_PL_ERR_RECOVERY);
#endif
}

u_short pltable_table_make( void )
{
  u_short         make_ret;
  u_long          pl_addr;
  u_int          *pl_ram_addr;
  short           flash_ret;
  u_int           r_size;
  char           *r_buff;

  make_ret = FALSE;

  pl_ram_addr = (u_int *)smem_alloc(SMEM_PLTABLE_INFO, 0x1000);
  if (pl_ram_addr == NULL)
  {
    printk(KERN_ERR "###PLTABLE pltable_table_make() 1 faild###\n");
    return(make_ret);
  }
  pltable = (PL_IMAGE_TABLE *)pl_ram_addr;
  memcpy(&pl_sw_table[0], (char *)pl_ram_addr+sizeof(PL_IMAGE_TABLE), sizeof(pl_sw_table));
  pl_sw_table[HS_PL_TABLE_A1].tbl   = &pltable->pl_table1[0];
  pl_sw_table[HS_PL_TABLE_A2].tbl   = &pltable->pl_table2[0];
  pl_sw_table[HS_PL_TABLE_A3].tbl   = &pltable->pl_table3[0];

  if ((pltable->status1 == HS_PL_MAKE_FLAG1) && (pltable->status2 == HS_PL_MAKE_FLAG2))
  {
    return( TRUE );
  }

  r_buff = (char *)kmalloc(FLASH_NAND_DEVICE_BLOCK_SIZE, GFP_KERNEL);
  if (!r_buff)
  {
    printk(KERN_ERR "###PLTABLE pltable_table_make() 2 faild###\n");
    return(make_ret);
  }


  for (pl_addr = HS_DEF1; pl_addr < HS_DEF2; pl_addr+=FLASH_NAND_DEVICE_BLOCK_SIZE)
  {
    flash_ret = mtd->read(mtd, pl_addr - HS_DEF1, FLASH_NAND_DEVICE_PAGE_MAX_SIZE, &r_size, r_buff);
    if (flash_ret == -EUCLEAN)
    {
      flash_ret = 0;
    }
    if (!flash_ret)
    {
      memcpy(pltable, r_buff, sizeof(PL_IMAGE_TABLE));
      if ((pltable->status1 == HS_PL_MAKE_FLAG1) && (pltable->status2 == HS_PL_MAKE_FLAG2))
      {
        make_ret = TRUE;
        break;
      }
    }
  }

  kfree(r_buff);

  return(make_ret);
}

u_long  pltable_get_real( u_long logi_blkno )
{
  u_short         cont, plsw;
  u_short         blk_no;
  u_long          physi_blkno;

  pltable_is_area(logi_blkno, &cont, &plsw);

  if (cont == HS_PL_CONT_2)
  {
    blk_no      = logi_blkno - pl_sw_table[plsw].bl_top;
    physi_blkno = pl_sw_table[plsw].tbl[blk_no];
  }
  else if (cont == HS_PL_CONT_1)
  {
    physi_blkno = logi_blkno;
  }
  else
  {
    printk(KERN_ERR "###PLTABLE pltable_get_real() faild###\n");
    physi_blkno = HS_PL_ERR_ADDRESS;
  }
#ifdef NO_PLTBL
  physi_blkno = logi_blkno;
#endif
  return(physi_blkno);

}

u_long  pltable_search_spare( u_long bad_block )
{
#if 1
  u_short         loop;
  u_short         cont, plsw;
  u_short         search_blk;
  u_short         data_n;
  u_short         s_blno, e_blno;
  u_short        *p_pltbl;
  u_long          chk_addr;
  short           flash_ret;

  pltable_is_area(bad_block, &cont, &plsw);

  if (cont == HS_PL_CONT_2)
  {
    p_pltbl = &pl_sw_table[plsw].tbl[0];
    s_blno  = pl_sw_table[plsw].bl_top;
    e_blno  = pl_sw_table[plsw].bl_end;
    data_n  = pl_sw_table[plsw].data;

    if (plsw == HS_PL_TABLE_A1)
    {
      mtd = get_mtd_device(NULL, 0);
    }
    else if (plsw == HS_PL_TABLE_A2)
    {
      mtd = get_mtd_device(NULL, 3);
    }
    else if (plsw == HS_PL_TABLE_A3)
    {
      mtd = get_mtd_device(NULL, 6);
    }
    else
    {
      mtd = get_mtd_device(NULL, 4);
    }
    if (IS_ERR(mtd))
    {
      return FALSE;
    }
  }
  else
  {
    printk(KERN_ERR "###PLTABLE pltable_search_spare() 1 faild###\n");
    return(HS_PL_ERR_RECOVERY);
  }

  for (search_blk = e_blno; search_blk >= (s_blno+data_n); search_blk--)
  {
    for (loop = 0; loop < data_n; loop++)
    {
      if (search_blk == p_pltbl[loop])
      {
        break;
      }
    }
    if (loop >= data_n)
    {
      chk_addr  = (search_blk - s_blno) * FLASH_NAND_DEVICE_BLOCK_SIZE;
      flash_ret = mtd->block_isbad(mtd, chk_addr);
      if (flash_ret == 0)
      {
        return(search_blk);
      }
    }
  }
  return(HS_PL_ERR_RECOVERY);
#else
  return(bad_block);
#endif
}

u_short pltable_put_real( u_long bad_block, u_long new_block )
{
#if 1
  u_short         loop, s_search, e_search;
  u_short         cont, plsw;
  u_short         chg_block;
  u_short        *p_pltbl;
  u_long          pl_addr;
  u_short         flash_ret;
  struct erase_info ei;
  u_int           write_size = 0;
  char           *pl_buff;

  s_search = HS_DEF1 / FLASH_NAND_DEVICE_BLOCK_SIZE;
  e_search = HS_DEF2 / FLASH_NAND_DEVICE_BLOCK_SIZE;

  mtd = get_mtd_device(NULL, 4);
  if (IS_ERR(mtd))
  {
    printk(KERN_ERR "###PLTABLE PLTable_table_setting() 1 faild###\n");
    return FALSE;
  }

  pltable_is_area(bad_block, &cont, &plsw);

  if (cont == HS_PL_CONT_2)
  {
    p_pltbl  = &pl_sw_table[plsw].tbl[0];
  }
  else
  {
    printk(KERN_ERR "###PLTABLE pltable_put_real() 1 faild###\n");
    return(FALSE);
  }

  pl_buff = (char *)kmalloc(FLASH_NAND_DEVICE_BLOCK_SIZE, GFP_KERNEL);
  if (!pl_buff)
  {
    printk(KERN_ERR "###PLTABLE pltable_search_spare() 2 faild###\n");
    return(FALSE);
  }

  chg_block = bad_block - pl_sw_table[plsw].bl_top;
  p_pltbl[chg_block] = new_block;

  for (loop = s_search; loop < e_search; loop++)
  {
    pl_addr = (loop - s_search) * FLASH_NAND_DEVICE_BLOCK_SIZE;
    flash_ret = mtd->block_isbad(mtd, pl_addr);
    if (flash_ret == 0)
    {
      memset(&ei, 0, sizeof(struct erase_info));
      ei.mtd  = mtd;
      ei.addr = pl_addr & ~(FLASH_NAND_DEVICE_BLOCK_SIZE-1);
      ei.len  = mtd->erasesize;
      flash_ret = mtd->erase(mtd, &ei);
      if (flash_ret == 0)
      {
        break;
      }
    }
  }
  memcpy(pl_buff, pltable, sizeof(PL_IMAGE_TABLE));
  for (; loop < e_search; loop++)
  {
    pl_addr = (loop - s_search) * FLASH_NAND_DEVICE_BLOCK_SIZE;
    flash_ret = mtd->block_isbad(mtd, pl_addr);
    if (flash_ret == 1)
    {
      continue;
    }
    flash_ret = mtd->write(mtd, pl_addr, FLASH_NAND_DEVICE_PAGE_MAX_SIZE, &write_size, pl_buff);
    if (flash_ret == 0)
    {
      kfree(pl_buff);
      return(TRUE);
    }
  }

  printk(KERN_ERR "###PLTABLE pltable_put_real() 2 faild###\n");
  kfree(pl_buff);
  return(FALSE);
#else
  return(TRUE);
#endif
}

u_short pltable_is_area( u_short blkno, u_short *cont, u_short *plsw )
{
  u_short            loop;
  u_short            check;
  u_short            s_blno, e_blno;

  *cont = HS_PL_ERR_AREAOVER;
  *plsw = HS_PL_ERR_AREAOVER;
  check = HS_PL_ERR_AREAOVER;

  for (loop = 0; loop < HS_PL_TABLE_END; loop++)
  {
    s_blno  = pl_sw_table[loop].bl_top;
    e_blno  = pl_sw_table[loop].bl_top + pl_sw_table[loop].data - 1;
    if ((s_blno <= blkno) && (blkno <= e_blno))
    {
      *cont = pl_sw_table[loop].cont;
      *plsw = loop;
      check = HS_PL_SUC_ROMAREA;
      break;
    }
  }

  return(check);
}
#endif
